@@ -28,21 +28,18 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <sys/param.h> /* for MAXHOSTNAMELEN */
 #include <string.h>
+#include <netdb.h>
 #include <ctype.h>
 #include <errno.h>
 
-#ifdef HAVE_ALLOCA_H
-#include <alloca.h>
-#endif
-
 #include "gmime-utils.h"
 #include "gmime-table-private.h"
 #include "gmime-part.h"
 #include "gmime-charset.h"
 #include "gmime-iconv.h"
 #include "gmime-iconv-utils.h"
-#include "unicode.h"
 
 #define d(x)
 #define w(x) x
@@ -73,8 +70,6 @@
 #define DATE_TOKEN_NON_TIMEZONE_NUMERIC (1 << 6)
 #define DATE_TOKEN_HAS_SIGN             (1 << 7)
 
-/* from gmime.c */
-extern int gmime_interfaces_utf8;
 
 static char *base64_alphabet =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
@@ -190,7 +185,11 @@ g_mime_utils_header_format_date (time_t time, int offset)
 	
 	time += ((offset / 100) * (60 * 60)) + (offset % 100) * 60;
 	
+#ifdef HAVE_GMTIME_R
+	gmtime_r (&time, &tm);
+#else
 	memcpy (&tm, gmtime (&time), sizeof (tm));
+#endif
 	
 	return g_strdup_printf ("%s, %02d %s %04d %02d:%02d:%02d %+05d",
 				tm_days[tm.tm_wday], tm.tm_mday,
@@ -316,7 +315,7 @@ get_wday (const unsigned char *in, unsigned int inlen)
 		return -1;
 	
 	for (wday = 0; wday < 7; wday++)
-		if (!strncasecmp (in, tm_days[wday], 3))
+		if (!g_strncasecmp (in, tm_days[wday], 3))
 			return wday;
 	
 	return -1;  /* unknown week day */
@@ -348,7 +347,7 @@ get_month (const unsigned char *in, unsigned int inlen)
 		return -1;
 	
 	for (i = 0; i < 12; i++)
-		if (!strncasecmp (in, tm_months[i], 3))
+		if (!g_strncasecmp (in, tm_months[i], 3))
 			return i;
 	
 	return -1;  /* unknown month */
@@ -504,9 +503,9 @@ parse_rfc822_date (struct _date_token *tokens, int *tzone)
 	}
 	
 	t = mktime (&tm);
-#if defined (HAVE_TIMEZONE)
+#if defined(HAVE_TIMEZONE)
 	t -= timezone;
-#elif defined (HAVE_TM_GMTOFF)
+#elif defined(HAVE_TM_GMTOFF)
 	t += tm.tm_gmtoff;
 #else
 #error Neither HAVE_TIMEZONE nor HAVE_TM_GMTOFF defined. Rerun autoheader, autoconf, etc.
@@ -540,7 +539,7 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 	int hour, min, sec, offset, n;
 	struct _date_token *token;
 	struct tm tm;
-	time_t t;
+	time_t time;
 	
 	memset ((void *) &tm, 0, sizeof (struct tm));
 	got_wday = got_month = got_tzone = FALSE;
@@ -622,12 +621,12 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 	}
 	
 	d(printf ("\n"));
-	
-	t = mktime (&tm);
-#if defined (HAVE_TIMEZONE)
-	t -= timezone;
-#elif defined (HAVE_TM_GMTOFF)
-	t += tm.tm_gmtoff;
+		
+	time = mktime (&tm);
+#if defined(HAVE_TIMEZONE)
+	time -= timezone;
+#elif defined(HAVE_TM_GMTOFF)
+	time += tm.tm_gmtoff;
 #else
 #error Neither HAVE_TIMEZONE nor HAVE_TM_GMTOFF defined. Rerun autoheader, autoconf, etc.
 #endif
@@ -635,12 +634,12 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 	/* t is now GMT of the time we want, but not offset by the timezone ... */
 	
 	/* this should convert the time to the GMT equiv time */
-	t -= ((offset / 100) * 60 * 60) + (offset % 100) * 60;
+	time -= ((offset / 100) * 60 * 60) + (offset % 100) * 60;
 	
 	if (tzone)
 		*tzone = offset;
 	
-	return t;
+	return time;
 }
 
 #if 0
@@ -723,15 +722,197 @@ g_mime_utils_header_decode_date (const char *in, int *saveoffset)
 }
 
 
-static void
-g_string_append_len (GString *out, const char *in, size_t len)
+/**
+ * g_mime_utils_generate_message_id:
+ * @fqdn: Fully qualified domain name
+ *
+ * Generates a unique Message-Id.
+ *
+ * Returns a unique string in an addr-spec format suitable for use as
+ * a Message-Id.
+ **/
+char *
+g_mime_utils_generate_message_id (const char *fqdn)
 {
-	char *buf;
+#ifdef G_THREADS_ENABLED
+	static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
+#define MUTEX_LOCK()   g_static_mutex_lock (&mutex)
+#define MUTEX_UNLOCK() g_static_mutex_unlock (&mutex)
+#else
+#define MUTEX_LOCK()
+#define MUTEX_UNLOCK()
+#endif
+	static unsigned int count = 0;
+	char host[MAXHOSTNAMELEN + 1];
+	struct hostent *h = NULL;
+	char *msgid;
 	
-	buf = alloca (len + 1);
-	strlcpy (buf, in, len + 1);
+	if (!fqdn) {
+		if (gethostname (host, sizeof (host)) == 0) {
+			h = gethostbyname (host);
+		} else {
+			host[0] = '\0';
+		}
+		
+		fqdn = h ? h->h_name : (*host ? host : "localhost.localdomain");
+	}
 	
-	g_string_append (out, buf);
+	MUTEX_LOCK ();
+	msgid = g_strdup_printf ("%ul.%ul.%ul@%s", time (NULL), getpid (), count++, fqdn);
+	MUTEX_UNLOCK ();
+	
+	return msgid;
+}
+
+/* external symbols from internet-address.c */
+extern void decode_lwsp (const char **in);
+extern char *decode_word (const char **in);
+extern char *decode_addrspec (const char **in);
+
+
+static char *
+decode_msgid (const char **in)
+{
+	const char *inptr = *in;
+	char *msgid = NULL;
+	
+	decode_lwsp (&inptr);
+	if (*inptr == '<') {
+		decode_lwsp (&inptr);
+		if ((msgid = decode_addrspec (&inptr))) {
+			decode_lwsp (&inptr);
+			if (*inptr != '>') {
+				w(g_warning ("Invalid msg-id; missing '>': %s", *in));
+			} else {
+				inptr++;
+			}
+			
+			*in = inptr;
+		} else {
+			w(g_warning ("Invalid msg-id; missing addr-spec: %s", *in));
+		}
+	} else {
+		w(g_warning ("Invalid msg-id; missing '<': %s", *in));
+	}
+	
+	return msgid;
+}
+
+
+/**
+ * g_mime_utils_decode_message_id:
+ * @message_id: string containing a message-id
+ *
+ * Decodes a msg-id as defined by rfc822.
+ *
+ * Returns the addr-spec portion of the msg-id.
+ **/
+char *
+g_mime_utils_decode_message_id (const char *message_id)
+{
+	g_return_val_if_fail (message_id != NULL, NULL);
+	
+	return decode_msgid (&message_id);
+}
+
+
+/**
+ * g_mime_references_decode:
+ * @text: string containing a list of msg-ids
+ *
+ * Decodes a list of msg-ids as in the References and/or In-Reply-To
+ * headers defined in rfc822.
+ *
+ * Returns a list of referenced msg-ids.
+ **/
+GMimeReferences *
+g_mime_references_decode (const char *text)
+{
+	GMimeReferences *refs, *tail, *ref;
+	const char *inptr = text;
+	char *word, *msgid;
+	
+	g_return_val_if_fail (text != NULL, NULL);
+	
+	refs = NULL;
+	tail = (GMimeReferences *) &refs;
+	
+	while (*inptr) {
+		decode_lwsp (&inptr);
+		if (*inptr == '<') {
+			/* looks like a msg-id */
+			if ((msgid = decode_msgid (&inptr))) {
+				ref = g_new (GMimeReferences, 1);
+				ref->next = NULL;
+				ref->msgid = msgid;
+				tail->next = ref;
+				tail = ref;
+			} else {
+				w(g_warning ("Invalid References header: %s", inptr));
+				break;
+			}
+		} else if (*inptr) {
+			/* looks like part of a phrase */
+			if ((word = decode_word (&inptr))) {
+				g_free (word);
+			} else {
+				w(g_warning ("Invalid References header: %s", inptr));
+				break;
+			}
+		}
+	}
+	
+	return refs;
+}
+
+
+/**
+ * g_mime_references_append:
+ * @refs: the address of a GMimeReferences list
+ * @msgid: a message-id string
+ *
+ * Appends a reference to msgid to the list of references.
+ **/
+void
+g_mime_references_append (GMimeReferences **refs, const char *msgid)
+{
+	GMimeReferences *ref;
+	
+	g_return_if_fail (refs != NULL);
+	g_return_if_fail (msgid != NULL);
+	
+	ref = (GMimeReferences *) refs;
+	while (ref->next)
+		ref = ref->next;
+	
+	ref->next = g_new (GMimeReferences, 1);
+	ref->next->msgid = g_strdup (msgid);
+	ref->next->next = NULL;
+}
+
+
+/**
+ * g_mime_references_clear:
+ * @refs: address of a GMimeReferences list
+ *
+ * Clears the GMimeReferences list and resets it to %NULL.
+ **/
+void
+g_mime_references_clear (GMimeReferences **refs)
+{
+	GMimeReferences *ref, *next;
+	
+	g_return_if_fail (refs != NULL);
+	
+	ref = *refs;
+	while (ref) {
+		next = ref->next;
+		g_free (ref->msgid);
+		g_free (ref);
+		ref = next;
+	}
+	
+	*refs = NULL;
 }
 
 
@@ -946,7 +1127,7 @@ g_mime_utils_unquote_string (char *string)
  * Determines if @text contains 8bit characters within the first @len
  * bytes.
  *
- * Returns TRUE if the text contains 8bit characters or FALSE
+ * Returns %TRUE if the text contains 8bit characters or %FALSE
  * otherwise.
  **/
 gboolean
@@ -973,7 +1154,7 @@ g_mime_utils_text_is_8bit (const unsigned char *text, size_t len)
  * Determines the best content encoding for the first @len bytes of
  * @text.
  *
- * Returns a GMimePartEncodingType that is determined to be the best
+ * Returns a #GMimePartEncodingType that is determined to be the best
  * encoding type for the specified block of text. ("best" in this
  * particular case means best compression)
  **/
@@ -1037,6 +1218,11 @@ rfc2047_decode_word (const unsigned char *in, size_t inlen)
 {
 	const register unsigned char *inptr;
 	const unsigned char *inend;
+	const char *charset;
+	unsigned char *buf;
+	char *charenc, *p;
+	size_t len;
+	iconv_t cd;
 	
 	inptr = in + 2;
 	inend = in + inlen - 2;
@@ -1054,13 +1240,13 @@ rfc2047_decode_word (const unsigned char *in, size_t inlen)
 		case 'B':
 		case 'b':
 			inptr += 2;
-			decoded = alloca (inend - inptr);
+			decoded = g_alloca (inend - inptr);
 			declen = g_mime_utils_base64_decode_step (inptr, inend - inptr, decoded, &state, &save);
 			break;
 		case 'Q':
 		case 'q':
 			inptr += 2;
-			decoded = alloca (inend - inptr);
+			decoded = g_alloca (inend - inptr);
 			declen = quoted_decode (inptr, inend - inptr, decoded);
 			
 			if (declen == -1) {
@@ -1073,57 +1259,47 @@ rfc2047_decode_word (const unsigned char *in, size_t inlen)
 			return NULL;
 		}
 		
-		if (gmime_interfaces_utf8) {
-			const char *charset;
-			unsigned char *buf;
-			char *charenc, *p;
-			size_t len;
-			iconv_t cd;
-			
-			len = (inptr - 3) - (in + 2);
-			charenc = alloca (len + 1);
-			memcpy (charenc, in + 2, len);
-			charenc[len] = '\0';
-			charset = charenc;
-			
-			/* rfc2231 updates rfc2047 encoded words...
-			 * The ABNF given in RFC 2047 for encoded-words is:
-			 *   encoded-word := "=?" charset "?" encoding "?" encoded-text "?="
-			 * This specification changes this ABNF to:
-			 *   encoded-word := "=?" charset ["*" language] "?" encoding "?" encoded-text "?="
-			 */
-			
-			/* trim off the 'language' part if it's there... */
-			p = strchr (charset, '*');
-			if (p)
-				*p = '\0';
-			
-			/* slight optimization */
-			if (!strcasecmp (charset, "UTF-8"))
-				return g_strndup (decoded, declen);
-			
-			cd = g_mime_iconv_open ("UTF-8", charset);
-			if (cd == (iconv_t) -1) {
-				w(g_warning ("Cannot convert from %s to UTF-8, header display may "
-					     "be corrupt: %s", charset, g_strerror (errno)));
-				charset = g_mime_charset_locale_name ();
-				cd = g_mime_iconv_open ("UTF-8", charset);
-				if (cd == (iconv_t) -1)
-					return NULL;
-			}
-			
-			buf = g_mime_iconv_strndup (cd, decoded, declen);
-			g_mime_iconv_close (cd);
-			
-			if (!buf) {
-				w(g_warning ("Failed to convert \"%.*s\" to UTF-8, display may be "
-					     "corrupt: %s", declen, decoded, g_strerror (errno)));
-			}
-			
-			return buf;
-		} else {
+		len = (inptr - 3) - (in + 2);
+		charenc = g_alloca (len + 1);
+		memcpy (charenc, in + 2, len);
+		charenc[len] = '\0';
+		charset = charenc;
+		
+		/* rfc2231 updates rfc2047 encoded words...
+		 * The ABNF given in RFC 2047 for encoded-words is:
+		 *   encoded-word := "=?" charset "?" encoding "?" encoded-text "?="
+		 * This specification changes this ABNF to:
+		 *   encoded-word := "=?" charset ["*" language] "?" encoding "?" encoded-text "?="
+		 */
+		
+		/* trim off the 'language' part if it's there... */
+		p = strchr (charset, '*');
+		if (p)
+			*p = '\0';
+		
+		/* slight optimization */
+		if (!g_strcasecmp (charset, "UTF-8"))
 			return g_strndup (decoded, declen);
+		
+		cd = g_mime_iconv_open ("UTF-8", charset);
+		if (cd == (iconv_t) -1) {
+			w(g_warning ("Cannot convert from %s to UTF-8, header display may "
+				     "be corrupt: %s", charset, g_strerror (errno)));
+			charset = g_mime_charset_locale_name ();
+			cd = g_mime_iconv_open ("UTF-8", charset);
+			if (cd == (iconv_t) -1)
+				return NULL;
 		}
+		
+		buf = g_mime_iconv_strndup (cd, decoded, declen);
+		g_mime_iconv_close (cd);
+		
+		if (!buf) {
+			w(g_warning ("Failed to convert \"%.*s\" to UTF-8, display may be "
+				     "corrupt: %s", declen, decoded, g_strerror (errno)));
+		}
+		
+		return buf;
 	}
 	
 	return NULL;
@@ -1287,27 +1463,26 @@ rfc2047_encode_word (GString *string, const unsigned char *word, size_t len,
 	int save = 0;
 	char encoding;
 	
-	if (gmime_interfaces_utf8) {
-		if (strcasecmp (charset, "UTF-8") != 0)
-			cd = g_mime_iconv_open (charset, "UTF-8");
-		
-		if (cd != (iconv_t) -1) {
-			uword = g_mime_iconv_strndup (cd, word, len);
-			g_mime_iconv_close (cd);
-		}
-		
-		if (uword) {
-			len = strlen (uword);
-			word = uword;
-		} else {
-			charset = "UTF-8";
-		}
+	
+	if (g_strcasecmp (charset, "UTF-8") != 0)
+		cd = g_mime_iconv_open (charset, "UTF-8");
+	
+	if (cd != (iconv_t) -1) {
+		uword = g_mime_iconv_strndup (cd, word, len);
+		g_mime_iconv_close (cd);
+	}
+	
+	if (uword) {
+		len = strlen (uword);
+		word = uword;
+	} else {
+		charset = "UTF-8";
 	}
 	
 	switch (g_mime_utils_best_encoding (word, len)) {
 	case GMIME_PART_ENCODING_BASE64:
 		enclen = BASE64_ENCODE_LEN (len);
-		encoded = alloca (enclen);
+		encoded = g_alloca (enclen);
 		
 		encoding = 'b';
 		
@@ -1322,7 +1497,7 @@ rfc2047_encode_word (GString *string, const unsigned char *word, size_t len,
 		break;
 	case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
 		enclen = QP_ENCODE_LEN (len);
-		encoded = alloca (enclen);
+		encoded = g_alloca (enclen);
 		
 		encoding = 'q';
 		
@@ -1362,12 +1537,8 @@ g_mime_utils_8bit_header_encode_phrase (const unsigned char *in)
 	
 	len = strlen (in);
 	
-	if (gmime_interfaces_utf8) {
-		charset = g_mime_charset_best (in, len);
-		charset = charset ? charset : "iso-8859-1";
-	} else {
-		charset = g_mime_charset_locale_name ();
-	}
+	charset = g_mime_charset_best (in, len);
+	charset = charset ? charset : "iso-8859-1";
 	
 	string = g_string_new ("");
 	
@@ -1418,29 +1589,20 @@ rfc2047_encode_phrase_get_words (const unsigned char *in)
 	
 	last = start = inptr = in;
 	while (inptr && *inptr) {
-		gboolean is_space;
-		unichar c;
+		const char *newinptr;
+		gunichar c;
 		
-		if (gmime_interfaces_utf8) {
-			const char *newinptr;
-			
-			newinptr = unicode_next_char (inptr);
-			c = unicode_get_char (inptr);
-			if (newinptr == NULL || !unichar_validate (c)) {
-				w(g_warning ("Invalid UTF-8 sequence encountered"));
-				inptr++;
-				continue;
-			}
-			
-			inptr = newinptr;
-			
-			is_space = unichar_isspace (c);
-		} else {
-			is_space = isspace ((int) *inptr);
-			c = *inptr++;
+		newinptr = g_utf8_next_char (inptr);
+		c = g_utf8_get_char (inptr);
+		if (newinptr == NULL || !g_unichar_validate (c)) {
+			w(g_warning ("Invalid UTF-8 sequence encountered"));
+			inptr++;
+			continue;
 		}
 		
-		if (is_space) {
+		inptr = newinptr;
+		
+		if (g_unichar_isspace (c)) {
 			if (count > 0) {
 				word = g_new (struct _phrase_word, 1);
 				word->next = NULL;
@@ -1581,7 +1743,7 @@ rfc2047_encode_phrase (const unsigned char *in)
 				start = word->start;
 			}
 			
-			if (word->encoding == 1 || !gmime_interfaces_utf8)
+			if (word->encoding == 1)
 				rfc2047_encode_word (out, start, len, "iso-8859-1", IS_PSAFE);
 			else
 				rfc2047_encode_word (out, start, len,
@@ -1628,8 +1790,8 @@ g_mime_utils_8bit_header_encode (const unsigned char *in)
  * @save: leftover bits that have not yet been encoded
  *
  * Base64 encodes the input stream to the output stream. Call this
- * when finished encoding data with g_mime_utils_base64_encode_step to
- * flush off the last little bit.
+ * when finished encoding data with #g_mime_utils_base64_encode_step
+ * to flush off the last little bit.
  *
  * Returns the number of bytes encoded.
  **/
@@ -1755,6 +1917,7 @@ g_mime_utils_base64_encode_step (const unsigned char *in, size_t inlen, unsigned
 	return (outptr - out);
 }
 
+
 /**
  * g_mime_utils_base64_decode_step:
  * @in: input stream
@@ -1828,7 +1991,7 @@ g_mime_utils_base64_decode_step (const unsigned char *in, size_t inlen, unsigned
  * @save: leftover bits that have not yet been encoded
  *
  * Uuencodes a chunk of data. Call this when finished encoding data
- * with g_mime_utils_uuencode_step to flush off the last little bit.
+ * with #g_mime_utils_uuencode_step to flush off the last little bit.
  *
  * Returns the number of bytes encoded.
  **/
@@ -1980,7 +2143,7 @@ g_mime_utils_uuencode_step (const unsigned char *in, size_t inlen, unsigned char
 /**
  * g_mime_utils_uudecode_step:
  * @in: input stream
- * @inlen: max length of data to decode (normally strlen (in) ??)
+ * @inlen: max length of data to decode
  * @out: output stream
  * @state: holds the number of bits that are stored in @save
  * @save: leftover bits that have not yet been decoded
@@ -2089,8 +2252,8 @@ g_mime_utils_uudecode_step (const unsigned char *in, size_t inlen, unsigned char
  * @save: leftover bits that have not yet been encoded
  *
  * Quoted-printable encodes a block of text. Call this when finished
- * encoding data with g_mime_utils_quoted_encode_step to flush off the
- * last little bit.
+ * encoding data with #g_mime_utils_quoted_encode_step to flush off
+ * the last little bit.
  *
  * Returns the number of bytes encoded.
  **/
